@@ -15,27 +15,48 @@ pub struct ApiFile {
 
 impl ApiFile {
     /// [f] returns [true] if it wants to stop going to the *children* of this subtree
-    pub fn visit_types<F: FnMut(&ApiType) -> bool>(&self, f: &mut F) {
+    pub fn visit_types<F: FnMut(&ApiType) -> bool>(
+        &self,
+        f: &mut F,
+        include_func_inputs: bool,
+        include_func_output: bool,
+    ) {
         for func in &self.funcs {
-            for field in &func.inputs {
-                field.ty.visit_types(f, self);
+            if include_func_inputs {
+                for field in &func.inputs {
+                    field.ty.visit_types(f, self);
+                }
+            }
+            if include_func_output {
+                func.output.visit_types(f, self);
             }
-            func.output.visit_types(f, self);
         }
     }
 
-    pub fn distinct_types(&self) -> Vec<ApiType> {
+    pub fn distinct_types(
+        &self,
+        include_func_inputs: bool,
+        include_func_output: bool,
+    ) -> Vec<ApiType> {
         let mut seen_idents = HashSet::new();
         let mut ans = Vec::new();
-        self.visit_types(&mut |ty| {
-            let ident = ty.safe_ident();
-            let contains = seen_idents.contains(&ident);
-            if !contains {
-                seen_idents.insert(ident);
-                ans.push(ty.clone());
-            }
-            contains
-        });
+        self.visit_types(
+            &mut |ty| {
+                let ident = ty.safe_ident();
+                let contains = seen_idents.contains(&ident);
+                if !contains {
+                    seen_idents.insert(ident);
+                    ans.push(ty.clone());
+                }
+                contains
+            },
+            include_func_inputs,
+            include_func_output,
+        );
+
+        // make the output change less when input change
+        ans.sort_by_key(|ty| ty.safe_ident());
+
         ans
     }
 }
@@ -106,6 +127,22 @@ pub enum ApiType {
     Boxed(Box<ApiTypeBoxed>),
 }
 
+macro_rules! api_type_call_child {
+    ($func:ident, $ret:ty) => {
+        pub fn $func(&self) -> $ret {
+            match self {
+                Primitive(inner) => inner.$func(),
+                Delegate(inner) => inner.$func(),
+                PrimitiveList(inner) => inner.$func(),
+                GeneralList(inner) => inner.$func(),
+                StructRef(inner) => inner.$func(),
+                Boxed(inner) => inner.$func(),
+                Optional(inner) => inner.$func(),
+            }
+        }
+    };
+}
+
 impl ApiType {
     pub fn visit_types<F: FnMut(&ApiType) -> bool>(&self, f: &mut F, api_file: &ApiFile) {
         if f(self) {
@@ -129,84 +166,13 @@ impl ApiType {
         }
     }
 
-    pub fn safe_ident(&self) -> String {
-        match self {
-            Primitive(inner) => inner.safe_ident(),
-            Delegate(inner) => inner.safe_ident(),
-            PrimitiveList(inner) => inner.safe_ident(),
-            GeneralList(inner) => inner.safe_ident(),
-            StructRef(inner) => inner.safe_ident(),
-            Boxed(inner) => inner.safe_ident(),
-            Optional(inner) => inner.safe_ident(),
-        }
-    }
-
-    pub fn dart_api_type(&self) -> String {
-        match self {
-            Primitive(inner) => inner.dart_api_type(),
-            Delegate(inner) => inner.dart_api_type(),
-            PrimitiveList(inner) => inner.dart_api_type(),
-            GeneralList(inner) => inner.dart_api_type(),
-            StructRef(inner) => inner.dart_api_type(),
-            Boxed(inner) => inner.dart_api_type(),
-            Optional(inner) => inner.dart_api_type(),
-        }
-    }
-    pub fn dart_wire_type(&self) -> String {
-        match self {
-            Primitive(inner) => inner.dart_wire_type(),
-            Delegate(inner) => inner.dart_wire_type(),
-            PrimitiveList(inner) => inner.dart_wire_type(),
-            GeneralList(inner) => inner.dart_wire_type(),
-            StructRef(inner) => inner.dart_wire_type(),
-            Boxed(inner) => inner.dart_wire_type(),
-            Optional(inner) => inner.dart_wire_type(),
-        }
-    }
-    pub fn rust_api_type(&self) -> String {
-        match self {
-            Primitive(inner) => inner.rust_api_type(),
-            Delegate(inner) => inner.rust_api_type(),
-            PrimitiveList(inner) => inner.rust_api_type(),
-            GeneralList(inner) => inner.rust_api_type(),
-            StructRef(inner) => inner.rust_api_type(),
-            Boxed(inner) => inner.rust_api_type(),
-            Optional(inner) => inner.rust_api_type(),
-        }
-    }
-    pub fn rust_wire_type(&self) -> String {
-        match self {
-            Primitive(inner) => inner.rust_wire_type(),
-            Delegate(inner) => inner.rust_wire_type(),
-            PrimitiveList(inner) => inner.rust_wire_type(),
-            GeneralList(inner) => inner.rust_wire_type(),
-            StructRef(inner) => inner.rust_wire_type(),
-            Boxed(inner) => inner.rust_wire_type(),
-            Optional(inner) => inner.rust_wire_type(),
-        }
-    }
-    pub fn rust_wire_modifier(&self) -> String {
-        match self {
-            Primitive(inner) => inner.rust_wire_modifier(),
-            Delegate(inner) => inner.rust_wire_modifier(),
-            PrimitiveList(inner) => inner.rust_wire_modifier(),
-            GeneralList(inner) => inner.rust_wire_modifier(),
-            StructRef(inner) => inner.rust_wire_modifier(),
-            Boxed(inner) => inner.rust_wire_modifier(),
-            Optional(inner) => inner.rust_wire_modifier(),
-        }
-    }
-    pub fn rust_wire_is_pointer(&self) -> bool {
-        match self {
-            Primitive(inner) => inner.rust_wire_is_pointer(),
-            Delegate(inner) => inner.rust_wire_is_pointer(),
-            PrimitiveList(inner) => inner.rust_wire_is_pointer(),
-            GeneralList(inner) => inner.rust_wire_is_pointer(),
-            StructRef(inner) => inner.rust_wire_is_pointer(),
-            Boxed(inner) => inner.rust_wire_is_pointer(),
-            Optional(inner) => inner.rust_wire_is_pointer(),
-        }
-    }
+    api_type_call_child!(safe_ident, String);
+    api_type_call_child!(dart_api_type, String);
+    api_type_call_child!(dart_wire_type, String);
+    api_type_call_child!(rust_api_type, String);
+    api_type_call_child!(rust_wire_type, String);
+    api_type_call_child!(rust_wire_modifier, String);
+    api_type_call_child!(rust_wire_is_pointer, bool);
 
     #[inline]
     pub fn required_modifier(&self) -> &'static str {
@@ -262,8 +228,13 @@ pub trait ApiTypeChild {
 pub enum ApiTypePrimitive {
     U8,
     I8,
+    U16,
+    I16,
+    U32,
     I32,
+    U64,
     I64,
+    F32,
     F64,
     Bool,
 }
@@ -275,11 +246,15 @@ impl ApiTypeChild for ApiTypePrimitive {
 
     fn dart_api_type(&self) -> String {
         match self {
-            ApiTypePrimitive::U8 => "int",
-            ApiTypePrimitive::I8 => "int",
-            ApiTypePrimitive::I32 => "int",
-            ApiTypePrimitive::I64 => "int",
-            ApiTypePrimitive::F64 => "double",
+            ApiTypePrimitive::U8
+            | ApiTypePrimitive::I8
+            | ApiTypePrimitive::U16
+            | ApiTypePrimitive::I16
+            | ApiTypePrimitive::U32
+            | ApiTypePrimitive::I32
+            | ApiTypePrimitive::U64
+            | ApiTypePrimitive::I64 => "int",
+            ApiTypePrimitive::F32 | ApiTypePrimitive::F64 => "double",
             ApiTypePrimitive::Bool => "bool",
         }
         .to_string()
@@ -297,8 +272,13 @@ impl ApiTypeChild for ApiTypePrimitive {
         match self {
             ApiTypePrimitive::U8 => "u8",
             ApiTypePrimitive::I8 => "i8",
+            ApiTypePrimitive::U16 => "u16",
+            ApiTypePrimitive::I16 => "i16",
+            ApiTypePrimitive::U32 => "u32",
             ApiTypePrimitive::I32 => "i32",
+            ApiTypePrimitive::U64 => "u64",
             ApiTypePrimitive::I64 => "i64",
+            ApiTypePrimitive::F32 => "f32",
             ApiTypePrimitive::F64 => "f64",
             ApiTypePrimitive::Bool => "bool",
         }
@@ -314,8 +294,13 @@ impl ApiTypePrimitive {
         match self {
             ApiTypePrimitive::U8 | ApiTypePrimitive::Bool => "ffi.Uint8",
             ApiTypePrimitive::I8 => "ffi.Int8",
+            ApiTypePrimitive::U16 => "ffi.Uint16",
+            ApiTypePrimitive::I16 => "ffi.Int16",
+            ApiTypePrimitive::U32 => "ffi.Uint32",
             ApiTypePrimitive::I32 => "ffi.Int32",
+            ApiTypePrimitive::U64 => "ffi.Uint64",
             ApiTypePrimitive::I64 => "ffi.Int64",
+            ApiTypePrimitive::F32 => "ffi.Float",
             ApiTypePrimitive::F64 => "ffi.Double",
         }
     }
@@ -323,8 +308,13 @@ impl ApiTypePrimitive {
         match s {
             "u8" => Some(ApiTypePrimitive::U8),
             "i8" => Some(ApiTypePrimitive::I8),
+            "u16" => Some(ApiTypePrimitive::U16),
+            "i16" => Some(ApiTypePrimitive::I16),
+            "u32" => Some(ApiTypePrimitive::U32),
             "i32" => Some(ApiTypePrimitive::I32),
+            "u64" => Some(ApiTypePrimitive::U64),
             "i64" => Some(ApiTypePrimitive::I64),
+            "f32" => Some(ApiTypePrimitive::F32),
             "f64" => Some(ApiTypePrimitive::F64),
             "bool" => Some(ApiTypePrimitive::Bool),
             _ => None,
@@ -336,8 +326,7 @@ impl ApiTypePrimitive {
 #[derive(Debug, Clone)]
 pub enum ApiTypeDelegate {
     String,
-    // upstream (allo-isolate) only supports ZeroCopyBuffer for Vec<u8> and Vec<i8>.
-    ZeroCopyBufferVecU8,
+    ZeroCopyBufferVecPrimitive(ApiTypePrimitive),
 }
 
 impl ApiTypeDelegate {
@@ -346,9 +335,11 @@ impl ApiTypeDelegate {
             ApiTypeDelegate::String => ApiType::PrimitiveList(ApiTypePrimitiveList {
                 primitive: ApiTypePrimitive::U8,
             }),
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => ApiType::PrimitiveList(ApiTypePrimitiveList {
-                primitive: ApiTypePrimitive::U8,
-            }),
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(primitive) => {
+                ApiType::PrimitiveList(ApiTypePrimitiveList {
+                    primitive: primitive.clone(),
+                })
+            }
         }
     }
 }
@@ -357,7 +348,7 @@ impl ApiTypeChild for ApiTypeDelegate {
     fn safe_ident(&self) -> String {
         match self {
             ApiTypeDelegate::String => "String".to_string(),
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => {
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => {
                 "ZeroCopyBuffer_".to_owned() + &self.get_delegate().dart_api_type()
             }
         }
@@ -366,7 +357,7 @@ impl ApiTypeChild for ApiTypeDelegate {
     fn dart_api_type(&self) -> String {
         match self {
             ApiTypeDelegate::String => "String".to_string(),
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => self.get_delegate().dart_api_type(),
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => self.get_delegate().dart_api_type(),
         }
     }
 
@@ -377,7 +368,9 @@ impl ApiTypeChild for ApiTypeDelegate {
     fn rust_api_type(&self) -> String {
         match self {
             ApiTypeDelegate::String => "String".to_owned(),
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => "ZeroCopyBuffer<Vec<u8>>".to_owned(),
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => {
+                format!("ZeroCopyBuffer<{}>", self.get_delegate().rust_api_type())
+            }
         }
     }
 
@@ -390,16 +383,6 @@ impl ApiTypeChild for ApiTypeDelegate {
     }
 }
 
-impl ApiTypeDelegate {
-    pub fn try_from_rust_str(s: &str) -> Option<Self> {
-        match s {
-            "String" => Some(ApiTypeDelegate::String),
-            "ZeroCopyBuffer<Vec<u8>>" => Some(ApiTypeDelegate::ZeroCopyBufferVecU8),
-            _ => None,
-        }
-    }
-}
-
 #[derive(Debug, Clone)]
 pub struct ApiTypePrimitiveList {
     pub primitive: ApiTypePrimitive,
@@ -414,7 +397,13 @@ impl ApiTypeChild for ApiTypePrimitiveList {
         match &self.primitive {
             ApiTypePrimitive::U8 => "Uint8List",
             ApiTypePrimitive::I8 => "Int8List",
+            ApiTypePrimitive::U16 => "Uint16List",
+            ApiTypePrimitive::I16 => "Int16List",
+            ApiTypePrimitive::U32 => "Uint32List",
+            ApiTypePrimitive::I32 => "Int32List",
+            ApiTypePrimitive::U64 => "Uint64List",
             ApiTypePrimitive::I64 => "Int64List",
+            ApiTypePrimitive::F32 => "Float32List",
             ApiTypePrimitive::F64 => "Float64List",
             _ => panic!("does not support {:?} yet", &self.primitive),
         }
