@@ -17,8 +17,11 @@ pub fn generate(
     dart_api_impl_class_name: &str,
     dart_wire_class_name: &str,
 ) -> Output {
-    let distinct_types = api_file.distinct_types();
-    debug!("distinct_types={:?}", distinct_types);
+    let distinct_types = api_file.distinct_types(true, true);
+    let distinct_input_types = api_file.distinct_types(true, false);
+    let distinct_output_types = api_file.distinct_types(false, true);
+    debug!("distinct_input_types={:?}", distinct_input_types);
+    debug!("distinct_output_types={:?}", distinct_output_types);
 
     let dart_func_signatures_and_implementations = api_file
         .funcs
@@ -36,15 +39,15 @@ pub fn generate(
         })
         .map(generate_api_struct)
         .collect::<Vec<_>>();
-    let dart_api2wire_funcs = distinct_types
+    let dart_api2wire_funcs = distinct_input_types
         .iter()
         .map(generate_api2wire_func)
         .collect::<Vec<_>>();
-    let dart_api_fill_to_wire_funcs = distinct_types
+    let dart_api_fill_to_wire_funcs = distinct_input_types
         .iter()
         .map(|ty| generate_api_fill_to_wire_func(ty, api_file))
         .collect::<Vec<_>>();
-    let dart_wire2api_funcs = distinct_types
+    let dart_wire2api_funcs = distinct_output_types
         .iter()
         .map(|ty| generate_wire2api_func(ty, api_file))
         .collect::<Vec<_>>();
@@ -198,14 +201,9 @@ fn generate_api2wire_func(ty: &ApiType) -> String {
             ApiTypeDelegate::String => {
                 "return _api2wire_uint_8_list(utf8.encoder.convert(raw));".to_string()
             }
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => {
-                "return _api2wire_uint_8_list(raw);".to_string()
-            } // ApiTypeDelegate::Optional(_) => {
-              // format!(
-              // "return raw != null ? _api2wire_{}(raw) : ffi.nullptr;",
-              // d.get_delegate().safe_ident()
-              // )
-              // }
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => {
+                format!("return _api2wire_{}(raw);", d.get_delegate().safe_ident())
+            }
         },
         Optional(opt) => format!(
             "return raw == null ? ffi.nullptr : _api2wire_{}(raw);",
@@ -323,7 +321,9 @@ fn generate_wire2api_func(ty: &ApiType, api_file: &ApiFile) -> String {
         Primitive(p) => gen_simple_type_cast(&p.dart_api_type()),
         Delegate(d) => match d {
             ApiTypeDelegate::String => gen_simple_type_cast(&d.dart_api_type()),
-            ApiTypeDelegate::ZeroCopyBufferVecU8 => gen_simple_type_cast(&d.dart_api_type()),
+            ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => {
+                gen_simple_type_cast(&d.dart_api_type())
+            }
         },
         Optional(opt) => format!(
             "return raw == null ? null : _wire2api_{}(raw);",
