@@ -11,9 +11,9 @@ pub struct Output {
     pub extern_func_names: Vec<String>,
 }
 
-pub fn generate(api_file: &ApiFile, rust_wire_stem: &str) -> Output {
+pub fn generate(api_file: &ApiFile, rust_wire_mod: &str) -> Output {
     let mut generator = Generator::new();
-    let code = generator.generate(api_file, rust_wire_stem);
+    let code = generator.generate(api_file, rust_wire_mod);
 
     Output {
         code,
@@ -32,34 +32,32 @@ impl Generator {
         }
     }
 
-    fn generate(&mut self, api_file: &ApiFile, rust_wire_stem: &str) -> String {
+    fn generate(&mut self, api_file: &ApiFile, rust_wire_mod: &str) -> String {
+        let distinct_input_types = api_file.distinct_types(true, false);
+        let distinct_output_types = api_file.distinct_types(false, true);
+
         let wire_funcs = api_file
             .funcs
             .iter()
             .map(|f| self.generate_wire_func(f))
             .collect::<Vec<_>>();
-        let wire_structs = api_file
-            .distinct_types()
+        let wire_structs = distinct_input_types
             .iter()
             .map(|ty| self.generate_wire_struct(ty, api_file))
             .collect::<Vec<_>>();
-        let allocate_funcs = api_file
-            .distinct_types()
+        let allocate_funcs = distinct_input_types
             .iter()
             .map(|f| self.generate_allocate_funcs(f))
             .collect::<Vec<_>>();
-        let wire2api_funcs = api_file
-            .distinct_types()
+        let wire2api_funcs = distinct_input_types
             .iter()
             .map(|ty| self.generate_wire2api_func(ty, api_file))
             .collect::<Vec<_>>();
-        let new_with_nullptr_funcs = api_file
-            .distinct_types()
+        let new_with_nullptr_funcs = distinct_input_types
             .iter()
             .map(|ty| self.generate_new_with_nullptr_func(ty, api_file))
             .collect::<Vec<_>>();
-        let impl_intodart = api_file
-            .distinct_types()
+        let impl_intodart = distinct_output_types
             .iter()
             .map(|ty| self.generate_impl_intodart(ty, api_file))
             .collect::<Vec<_>>();
@@ -89,6 +87,19 @@ impl Generator {
             fn wire2api(self) -> T;
         }}
 
+        impl<T, S> Wire2Api<Option<T>> for *mut S
+        where
+            *mut S: Wire2Api<T>
+        {{
+            fn wire2api(self) -> Option<T> {{
+                if self.is_null() {{
+                    None
+                }} else {{
+                    Some(self.wire2api())
+                }}
+            }}
+        }}
+
         {}
 
         // Section: impl NewWithNullPtr
@@ -98,7 +109,7 @@ impl Generator {
         }}
 
         impl<T> NewWithNullPtr for *mut T {{
-            fn new_with_null_ptr() -> Self {{ 
+            fn new_with_null_ptr() -> Self {{
                 std::ptr::null_mut()
             }}
         }}
@@ -113,7 +124,7 @@ impl Generator {
 
         "#,
             CODE_HEADER,
-            rust_wire_stem,
+            rust_wire_mod,
             wire_funcs.join("\n\n"),
             wire_structs.join("\n\n"),
             allocate_funcs.join("\n\n"),
@@ -264,8 +275,8 @@ impl Generator {
                 &format!("new_{}", ty.safe_ident()),
                 &["len: i32"],
                 Some(&[
-                     list.rust_wire_modifier().as_str(),
-                     list.rust_wire_type().as_str()
+                    list.rust_wire_modifier().as_str(),
+                    list.rust_wire_type().as_str()
                 ].concat()),
                 &format!(
                     "let wrap = {} {{ ptr: support::new_leak_vec_ptr(<{}{}>::new_with_null_ptr(), len), len }};
@@ -283,7 +294,7 @@ impl Generator {
                             &format!("new_{}", ty.safe_ident()),
                             &[&format!("value: {}", prim.rust_wire_type())],
                             Some(&format!("*mut {}", prim.rust_wire_type())),
-                            "support::new_leak_box_ptr(value)"
+                            "support::new_leak_box_ptr(value)",
                         )
                     }
                     inner => {
@@ -294,11 +305,11 @@ impl Generator {
                             &format!(
                                 "support::new_leak_box_ptr({}::new_with_null_ptr())",
                                 inner.rust_wire_type()
-                            )
+                            ),
                         )
                     }
                 }
-            },
+            }
         }
     }
 
@@ -310,7 +321,9 @@ impl Generator {
                 ApiTypeDelegate::String => "let vec: Vec<u8> = self.wire2api();
                 String::from_utf8_lossy(&vec).into_owned()"
                     .into(),
-                ApiTypeDelegate::ZeroCopyBufferVecU8 => "ZeroCopyBuffer(self.wire2api())".into(),
+                ApiTypeDelegate::ZeroCopyBufferVecPrimitive(_) => {
+                    "ZeroCopyBuffer(self.wire2api())".into()
+                }
             },
             PrimitiveList(_) => "unsafe {
                 let wrap = support::box_from_leak_ptr(self);
@@ -354,8 +367,8 @@ impl Generator {
                 }
                 .into()
             }
-            // implicit delegation
-            Optional(_) => "if self.is_null() { None } else { Some(self.wire2api()) }".into(),
+            // handled by common impl
+            Optional(_) => return String::new(),
         };
 
         format!(
@@ -438,13 +451,15 @@ impl Generator {
 
         format!(
             "impl support::IntoDart for {} {{
-            fn into_dart(self) -> support::DartCObject {{
-                vec![
-                    {}
-                ].into_dart()
+                fn into_dart(self) -> support::DartCObject {{
+                    vec![
+                        {}
+                    ].into_dart()
+                }}
             }}
-        }}",
-            s.name, body,
+            impl support::IntoDartExceptPrimitive for {} {{}}
+            ",
+            s.name, body, s.name,
         )
     }
 }
